inline MyInterval radiusXYZn(MyInterval x, MyInterval y, MyInterval z, int exp)
{
    return power(sqrt(power(x, 2) + power(y, 2) + power(z, 2)), exp);
}

void setMinMax(MyInterval x, MyInterval y, MyInterval z,
               MyInterval &qx, MyInterval &qy, MyInterval &qz)
{
    qx = intervalHull(x, qx);
    qy = intervalHull(y, qy);
    qz = intervalHull(z, qz);
}

void xydivr52D(const MyInterval &x, const MyInterval &y, const MyInterval &z,
               MyInterval &tempxy, MyInterval &tempxz, MyInterval &tempyz)
// tempxy = xy/r^{5}
// tempxz = xz/r^{5}
// tempyz = yz/r^{5}
{
    MyInterval rMinus5 = radiusXYZn(x, y, z, -5);
    tempxy = rMinus5 * x * y;
    tempxz = rMinus5 * x * z;
    tempyz = rMinus5 * y * z;
}

void tempValues(MyInterval x, MyInterval y, MyInterval z,
                MyInterval &tempx, MyInterval &tempy, MyInterval &tempz,
                int rExp, int bExp)
// tempx = x^bExp/r^rExp
// tempy = y^bExp/r^rExp
// tempz = z^bExp/r^rExp
{
    MyInterval rMinusrExp = radiusXYZn(x, y, z, -rExp);
    tempx = rMinusrExp * power(x, bExp);
    tempy = rMinusrExp * power(y, bExp);
    tempz = rMinusrExp * power(z, bExp);
}

/*===========================================================================================*/

inline bool intersectSurfaceWithEdge(MyInterval coef, MyInterval qa, MyInterval qb, MyInterval qc,
                                     MyInterval &pointa)
// check weather qa has non-empty intersection with  coef* sqrt{qb + qc}
// qa is an interval, qb, qc are squared points
// the original equation is x = +/- coef * sqrt{y^2 + z^2}
// on input we have qb = y^2, qc = z^2
{
    return myIntersection(coef * sqrt(qb + qc), qa, pointa);
}

inline bool intersectLineWithEdge(MyInterval coef, MyInterval qa, MyInterval qb,
                                  MyInterval &pointa)
// check weather qa has non-empty intersection with  coef* qb
{
    return myIntersection(coef * qb, qa, pointa);
}

/*-------------------------------------------------------------------------------------------*/

void computeEightPoints(MyInterval coef, const MyInterval &x, const MyInterval &y, const MyInterval &z,
                        MyInterval &qx, MyInterval &qy, MyInterval &qz,
                        int rExp, int bExp)
// 8 points:
// intersection of the surface `x = +/-coef * \sqrt{y^2 + z^2}` and the edges of the qube `q`:
{
    MyInterval tempx, tempy, tempz;
    MyInterval xI;

    MyInterval yAcc = y.right();
    MyInterval yPower = power(yAcc, 2);
    MyInterval zAcc = z.right();
    MyInterval zPower = power(zAcc, 2); // y -- right, z -- right

    if (intersectSurfaceWithEdge(coef, x, yPower, zPower, xI))
    {
        tempValues(xI, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
        setMinMax(tempx, tempy, tempz, qx, qy, qz);
    }
    if (intersectSurfaceWithEdge(-coef, x, yPower, zPower, xI))
    {
        tempValues(xI, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
        setMinMax(tempx, tempy, tempz, qx, qy, qz);
    }

    zAcc = z.left(); // y -- right, z -- left
    zPower = power(zAcc, 2);
    if (intersectSurfaceWithEdge(coef, x, yPower, zPower, xI))
    {
        tempValues(xI, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
        setMinMax(tempx, tempy, tempz, qx, qy, qz);
    }
    if (intersectSurfaceWithEdge(-coef, x, yPower, zPower, xI))
    {
        tempValues(xI, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
        setMinMax(tempx, tempy, tempz, qx, qy, qz);
    }

    yAcc = y.left();
    yPower = power(yAcc, 2); // y -- left, z -- left
    if (intersectSurfaceWithEdge(coef, x, yPower, zPower, xI))
    {
        tempValues(xI, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
        setMinMax(tempx, tempy, tempz, qx, qy, qz);
    }
    if (intersectSurfaceWithEdge(-coef, x, yPower, zPower, xI))
    {
        tempValues(xI, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
        setMinMax(tempx, tempy, tempz, qx, qy, qz);
    }

    zAcc = z.right();
    //tutaj mozna ciut zoptymalizowac i nie liczyc ponownie zPower!
    zPower = power(zAcc, 2); // y -- left, z -- right
    if (intersectSurfaceWithEdge(coef, x, yPower, zPower, xI))
    {
        tempValues(xI, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
        setMinMax(tempx, tempy, tempz, qx, qy, qz);
    }
    if (intersectSurfaceWithEdge(-coef, x, yPower, zPower, xI))
    {
        tempValues(xI, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
        setMinMax(tempx, tempy, tempz, qx, qy, qz);
    }
}

/*-------------------------------------------------------------------------------------------*/

void computeFourPoints(MyInterval coef, const MyInterval &a, const MyInterval &c,
                       MyInterval &qa, MyInterval &qb, MyInterval &qc,
                       int rExp, int bExp)
// on input of this function we have z (we assume y = 0)
// we compute +/- coef * z_{L, R}, where z_L -- z.left(), z_R -- z.right()
{
    MyInterval tempa, tempb, tempc;
    MyInterval aI;
    MyInterval acc = c.right();

    if (intersectLineWithEdge(coef, a, acc, aI)) // (1)  +coef * c_R
    {
        tempValues(aI, 0.0, acc, tempa, tempb, tempc, rExp, bExp);
        setMinMax(tempa, tempb, tempc, qa, qb, qc);
    }
    if (intersectLineWithEdge(-coef, a, acc, aI)) // (2)  -coef * c_R
    {
        tempValues(aI, 0.0, acc, tempa, tempb, tempc, rExp, bExp);
        setMinMax(tempa, tempb, tempc, qa, qb, qc);
    }

    acc = c.left();
    if (intersectLineWithEdge(coef, a, acc, aI)) // (3)  coef * c_L
    {
        tempValues(aI, 0.0, acc, tempa, tempb, tempc, rExp, bExp);
        setMinMax(tempa, tempb, tempc, qa, qb, qc);
    }
    if (intersectLineWithEdge(-coef, a, acc, aI)) // (4)  -coef * c_L
    {
        tempValues(aI, 0.0, acc, tempa, tempb, tempc, rExp, bExp);
        setMinMax(tempa, tempb, tempc, qa, qb, qc);
    }
}

/*-------------------------------------------------------------------------------------------*/

void computeTwoPoints(const MyInterval &a,
                      MyInterval &qa, MyInterval &qb, MyInterval &qc,
                      int rExp, int bExp)
// on input of this function we have x = a (we assume y = 0, z = 0)
// we compute in 2 points: (a_{L/R}, 0, 0)
{
    MyInterval tempa, tempb, tempc;
    MyInterval acc;

    acc = a.left();
    tempValues(acc, 0.0, 0.0, tempa, tempb, tempc, rExp, bExp);
    setMinMax(tempa, tempb, tempc, qa, qb, qc);

    acc = a.right();
    tempValues(acc, 0.0, 0.0, tempa, tempb, tempc, rExp, bExp);
    setMinMax(tempa, tempb, tempc, qa, qb, qc);
} // computeTwoPoints

/*-------------------------------------------------------------------------------------------*/

void computeFourSimplePoints(const MyInterval &a, const MyInterval &c,
                             MyInterval &qa, MyInterval &qb, MyInterval &qc,
                             int rExp, int bExp)
// compute in 4 points: (a_{L/R}, 0, c_{L/R})
{
    MyInterval tempa, tempb, tempc;
    MyInterval aAcc, cAcc;

    aAcc = a.left();
    cAcc = c.left();
    tempValues(aAcc, 0.0, cAcc, tempa, tempb, tempc, rExp, bExp);
    setMinMax(tempa, tempb, tempc, qa, qb, qc);

    cAcc = c.right();
    tempValues(aAcc, 0.0, cAcc, tempa, tempb, tempc, rExp, bExp);
    setMinMax(tempa, tempb, tempc, qa, qb, qc);

    aAcc = a.right();
    tempValues(aAcc, 0.0, cAcc, tempa, tempb, tempc, rExp, bExp);
    setMinMax(tempa, tempb, tempc, qa, qb, qc);

    cAcc = c.left();
    tempValues(aAcc, 0.0, cAcc, tempa, tempb, tempc, rExp, bExp);
    setMinMax(tempa, tempb, tempc, qa, qb, qc);
} // computeFourSimplePoints

/*-------------------------------------------------------------------------------------------*/

void computeInOtherSpecificPoints(const MyInterval &coef, const MyInterval &x, const MyInterval &y, const MyInterval &z,
                                  MyInterval &qx, MyInterval &qy, MyInterval &qz,
                                  int rExp, int bExp)
// On output: {qx,qy,qz}=IntervalHull(x^bExp/r^{rExp}, y^bExp/r^{rExp}, z^bExp/r^{rExp})
// on corners of box x \times y \times z
{

    MyInterval tempx, tempy, tempz;
    MyInterval xAcc, yAcc, zAcc;

    // intersection of the surface `y = coef * \sqrt{x^2 + z^2}` and the edges of the qube `q`:
    computeEightPoints(coef, x, y, z, qx, qy, qz, rExp, bExp);

    // we have to compute 4 points when y = 0:
    // (+/- z_{L/R} * coef, 0, z_{L/R}), where z_L -- z.left(), z_R -- z.right()
    // and 4 points when z = 0:
    // (+/- y_{L/R} * coef, y_{L/R}, 0), where y_L -- y.left(), y_R -- y.right()

    MyInterval xI, yI, zI;
    // 4 points: (+/-z_{left/right}*coef, 0, +/-z_{left/right})
    // intersection of the line `x = +/-coef * z && y = 0` and the edges of the qube `q`:
    if (isSingular(y))
    {
        // on input of this function we have z, we assume y = 0
        // we compute +/- coef * z_{L, R}, where z_L -- z.left(), z_R -- z.right()
        computeFourPoints(coef, x, z, qx, qy, qz, rExp, bExp);

        // x -- any, y = 0, z -- any
        computeFourSimplePoints(x, z, qx, qy, qz, rExp, bExp);
    }

    // 4 points: (+/-z_{left/right}*coef, 0, +/-z_{left/right})
    // intersection of the line `x = coef * y && z = 0` and the edges of the qube `q`:
    if (isSingular(z))
    {
        // on input of this function we have y, we assume z = 0
        // we compute +/- coef * y_{L, R}, where y_L -- y.left(), y_R -- y.right()
        computeFourPoints(coef, x, y, qx, qz, qy, rExp, bExp);

        // x -- any, y -- any, z = 0
        computeFourSimplePoints(x, y, qx, qz, qy, rExp, bExp);
    }

    // x != 0, y = 0, z = 0
    if (!isSingular(x) && isSingular(y) && isSingular(z))
        computeTwoPoints(x, qx, qy, qz, rExp, bExp);
}

void computeInCorners(const MyInterval &x, const MyInterval &y, const MyInterval &z,
                      MyInterval &qx, MyInterval &qy, MyInterval &qz,
                      int rExp, int bExp)
// On output: {qx,qy,qz}=IntervalHull(x^bExp/r^{rExp}, y^bExp/r^{rExp}, z^bExp/r^{rExp})
// on corners of box x \times y \times z
//
//       3-------------4
//      /|            /|
//     / |           / |
//    /  8----------/--7
//   2-------------5  /
//   | /           | /
//   |/            |/
//   1-------------6
//
{
    MyInterval xAcc = x.left();
    MyInterval yAcc = y.left();
    MyInterval zAcc = z.left();
    // bottom-front-left corner: (1) x -- left, y -- left, z -- left
    MyInterval tempx, tempy, tempz;

    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    qx = tempx;
    qy = tempy;
    qz = tempz;

    // top-front-left corner: (2)  x -- left, y -- right, z -- left
    yAcc = y.right();
    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    setMinMax(tempx, tempy, tempz, qx, qy, qz);

    // (3)   x -- left, y -- right, z -- right
    zAcc = z.right();
    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    setMinMax(tempx, tempy, tempz, qx, qy, qz);

    // top-back-right corner: (4)  x -- right, y -- right, z -- right
    xAcc = x.right();
    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    setMinMax(tempx, tempy, tempz, qx, qy, qz);

    // top-front-right corner: (5)   x -- right, y -- right, z -- left
    zAcc = z.left();
    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    setMinMax(tempx, tempy, tempz, qx, qy, qz);

    // bottom-front-right corner: (6)    x -- right, y -- left, z -- left
    yAcc = y.left();
    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    setMinMax(tempx, tempy, tempz, qx, qy, qz);

    // bottom-right corner: (7)     x -- right, y -- left, z -- right
    zAcc = z.right();
    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    setMinMax(tempx, tempy, tempz, qx, qy, qz);

    // (8)     x -- left, y -- left, z -- right
    xAcc = x.leftBound();
    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    setMinMax(tempx, tempy, tempz, qx, qy, qz);
}

/*-------------------------------------------------------------------------------------------*/

void cuteCalculation(const MyInterval &x, const MyInterval &y, const MyInterval &z,
                     MyInterval &qx, MyInterval &qy, MyInterval &qz,
                     int rExp, int bExp)
/* in the paper: bExp = a, rExp = b
 * we are calculating: q^bExp/|q|^{rExp}
 * qx = x^bExp/r^{rExp}, qy = y^bExp/r^{rExp}, qz = z^bExp/r^{rExp}
 * we consider sufraces:
 * * x = +/- sqrt((y^2 + z^2)a/(b-a))
 * * y = +/- sqrt((x^2 + z^2)a/(b-a))
 * * z = +/- sqrt((x^2 + y^2)a/(b-a))
 * * x = 0
 * * y = 0
 * * z = 0
 */
{
    if (rExp <= bExp)
    {
        cout << "Wrong exponents!";
        exit(2);
    }
    MyInterval r = sqrt(power(x, 2) + power(y, 2) + power(z, 2));
    MyInterval r3 = power(r, -rExp);
    if (!testIfDoCute(r)) // we do a naive evaluation
    {
        qx = power(x, bExp) * r3;
        qy = power(y, bExp) * r3;
        qz = power(z, bExp) * r3;
        return;
    }

    // coef = a/(b - a)
    // first check the corners of the `qi` qube:
    computeInCorners(x, y, z, qx, qy, qz, rExp, bExp);
    // now we have: qx, qy, qz -- min and max from the corners

    MyInterval coef = sqrt(MyInterval(bExp) / (rExp - bExp));

    computeInOtherSpecificPoints(coef, x, y, z, qx, qy, qz, rExp, bExp);
    computeInOtherSpecificPoints(coef, y, x, z, qy, qx, qz, rExp, bExp);
    computeInOtherSpecificPoints(coef, z, x, y, qz, qx, qy, rExp, bExp);
}
