#ifndef estimatesfunctions_h
#define estimatesfunctions_h
#include "defo.h"

#include "bodies.h"

bool findCluster(const Bodies &bodies, int selBody, MyDouble delta, vector<int> &cluster);
/**
We look for the cluster containg the body <selBody>
Input:
  bodies - box in configuration space containing bodies  * Numbering of bodies is zero-based (0, 1, ..., bodies.N - 1).
  num - the first body in the cluster
  delta - the body is in the cluster, if there exists a chain of bodies connecting it to <selBody>
     and each distance in the chain is less than <delta>
On output:
      In the `cluster` vector we have indices of the bodies from the cluster.
 * Returns `false` if the cluster is a singleton (i.e. the <selBody>-th body is isolated,
 * meaning that there're no other bodies within the delta radius).
 */

MyInterval supMomInertiaCluster(const Bodies &bodies, const vector<int> &cluster);
/// Returns  I_{C,Z} \geq \sum_{i\in cluster} m_i * q_i^2

MyInterval infPotEnergyCluster(const Bodies &bodies, const vector<int> &cluster);
// returns U_{C,Z} = inf \sum_{i < j\in cluster} \frac{m_i*m_j}{r_{i, j}}

MyInterval infFCZ(const Bodies &bodies, const vector<int> &cluster);
/*  F_{C,Z}(q) = \sum_{i\in cluster, k\not\in cluster} \frac{m_i * m_k}{r_{i, k}^3} (q_i - q_k|q_i)
 * */

bool isInCluster(const vector<int> &cluster, int i);

void updateVectorBodiesInCluster(vector<int> &bodiesInCluster, const vector<int> &cluster);
//  for each i in <cluster>
//   the function sets bodiesInCluster[i]=1;

#endif
